@@ -26,6 +26,7 @@
 #include <math.h>
 #include <time.h>
 #include <sys/stat.h>
+#include <pwd.h>
 
 #include "gkrellmwho2.h"
 
@@ -42,6 +43,8 @@ static GtkWidget	*away_spin_button;
 static GtkWidget	*gone_spin_button;
 static GtkWidget	*color_dialog = NULL;
 static GtkWidget	*status_color_str[3];
+static GtkWidget	*notify_once_button;
+static GtkWidget	*notify_command_str;
 
 static GkrellmMonitor	*monitor;
 static GkrellmPanel		*panel;
@@ -115,26 +118,13 @@ static void format_user_str(int user, char *str)
 	}
 }
 
-static void clear_users()
+static gint find_user( WhoUser *ua[], char *name )
 {
-	while( user_count-- > 0 ) {
-		free(users[user_count]->name);
-		g_free(users[user_count]->usr_str);
-		g_free(users[user_count]->text_style);
-		free(users[user_count]);
-	}
-	
-    user_count = 0;
-}
+	gint	i = 0;
 
-static gint find_user( char *name )
-{
-	gint	i;
-
-	for( i = 0; i < user_count; i++ ) {
-		if( !strcmp(users[i]->name, name) )
-			return i;
-	}
+	while( ua[i] )
+		if( !strcmp(ua[i++]->name, name) )
+			return i - 1;
 
 	return -1;
 }
@@ -146,7 +136,7 @@ static int	add_user( char *name, time_t idle_time )
 #ifdef DEBUG
 		g_print( "gkrellmwho2: adding user (%s) with idle time (%d)\n", name, (int)idle_time );
 #endif
-	if(	(uid = find_user(name)) < 0 ) {
+	if(	(uid = find_user(users, name)) < 0 ) {
 		/* We have a new user here */
 		users[user_count] = malloc(sizeof(WhoUser));
 		users[user_count]->name = strdup(name);
@@ -162,6 +152,35 @@ static int	add_user( char *name, time_t idle_time )
 	return 0;
 }
 
+static int	cache_users()
+{
+	gint	i;
+
+#ifdef DEBUG
+	g_print( "user_count: %d\n", user_count );
+#endif
+	for( i = 0; i < user_count; i++ ) {
+		if( cache[i] ) {
+			free( cache[i]->name );
+			g_free( cache[i]->usr_str );
+			g_free( cache[i]->text_style );
+			free( cache[i] );
+			cache[i] = NULL;
+		}
+		cache[i] = users[i];
+		users[i] = NULL;
+	}
+
+    user_count = 0;
+	return 0;
+}
+
+static void	status_change_notify( WhoUser *current, WhoUser *old )
+{
+	if( notify_once && notify_ran++ ) return;
+	g_spawn_command_line_async( notify_command, NULL );
+}
+
 static int	sort_idlestatus( WhoUser **a, WhoUser **b )
 {
 	return idle_status((*a)->idle_time) - idle_status((*b)->idle_time);
@@ -181,17 +200,21 @@ static void	get_users()
 {
 	FILE	*p;
 	time_t	idle_time = 0;
-	int		i;
+	int		i, uid, idle;
 	int		pos = 0;
 	int		w[4] = {0,0,0,0};
 	struct stat sbuf;
+	struct passwd *pw;
 	char	*str, *cur_str;
 	char	user[MAX_STRING_LEN];
 	char	line[MAX_STRING_LEN];
 	char	tmp[MAX_STRING_LEN];
 	char	host[MAX_STRING_LEN];
+	sortfn	*fn;
 
-	clear_users();
+	/*clear_users( users );*/
+	cache_users();
+	notify_ran = 0;
 	
 	if( !( p=popen( COMMAND, "r" )) ) {
 		/* strcpy( scroll_text, "COULD NOT OPEN PIPE" ); */
@@ -209,8 +232,8 @@ static void	get_users()
     while( !feof( p ) && user_count<MAX_USERS )  {
 		user[0] = '\0';
 		/* get user name */
-		fscanf( p, "%[^\n]\n", cur_str );
-		sscanf( cur_str, "%9s %8s %16s %*[^\n]",
+		fscanf( p, "%255[^\n]\n", cur_str );
+		sscanf( cur_str, "%8s %8s %16s %*[^\n]",
 				user, line, host );
 
 		if( !user[0] ) break; /* if COMMAND returns no users */
@@ -225,11 +248,15 @@ static void	get_users()
 			
 			/* Get user's tty idle time in unix format */
 			if( snprintf( tmp, MAX_STRING_LEN, "%s%s", DEVTTY, line ) &&
-				stat(tmp, &sbuf) == 0 )
-					idle_time = time(NULL) - sbuf.st_atime;
+				stat(tmp, &sbuf) == 0 ) {
+				idle_time = time(NULL) - sbuf.st_atime;
+				pw = getpwuid(sbuf.st_uid);
+				if( strlen(pw->pw_name) > 8 ) /* FIXME */
+					strcpy(user, pw->pw_name);
+			}
 
 			/* add it only if it's not there already */
-			add_user(user, idle_time);
+			add_user( user, idle_time );
 		}
     }
     pclose( p );
@@ -243,31 +270,42 @@ static void	get_users()
 
 		/* Colorize users */
 		for( i = 0; i < user_count; i++) {
-			w[idle_status(users[i]->idle_time)+1]++;
+			idle = idle_status( users[i]->idle_time );
+			if( ((uid = find_user( cache, users[i]->name )) >= 0 &&
+				( idle != idle_status(cache[uid]->idle_time) )) ||
+				(uid < 0) )
+				status_change_notify( users[i], cache[uid] );
+				/*g_print( "####################\n#\n#\n#\tUser %s changed it's status!\n"
+						"# %d -> %d\n#\n################\n", users[i]->name, users[i]->idle_time, 
+						cache[uid] ? cache[uid]->idle_time : 0 ); */
+			w[idle + 1]++;
 			users[i]->text_style = gkrellm_copy_textstyle(ts);
 			users[i]->text_style->color =
-				status_colors[ idle_status( users[i]->idle_time ) ];
+				status_colors[idle];
 		}
 
 		/* Let's sort users right here */
+		switch( sort_by ) {
+			case by_name:
+				fn = (sortfn *) sort_name;
+				break;
+			case by_idle:
+				fn = (sortfn *) sort_idle;
+				break;
+			default:
+				fn = (sortfn *) sort_name;
+		}
+
 		/* First, by idle status */
 		if( group_by_status ) {
 			qsort( users, user_count, sizeof(WhoUser *), (sortfn *)sort_idlestatus );
-			if( sort_by == by_name )
-				for( i = 0; i < 3; i++ ) {
-					qsort( &users[pos], w[i+1], sizeof(WhoUser *), (sortfn *)sort_name );
-					pos += w[i+1];
-				}
-			else if( sort_by == by_idle )
-				for( i = 0; i < 3; i++ ) {
-					qsort( &users[pos], w[i+1], sizeof(WhoUser *), (sortfn *)sort_idle );
-					pos += w[i+1];
-				}
-		} else if( sort_by == by_name ) {
-			qsort( users, user_count, sizeof(WhoUser *), (sortfn *)sort_name );
-		} else if( sort_by == by_idle ) {
-			qsort( users, user_count, sizeof(WhoUser *), (sortfn *)sort_idle );
-		}
+
+			for( i = 0; i < 3; i++ ) {
+				qsort( &users[pos], w[i+1], sizeof(WhoUser *), fn );
+				pos += w[i+1];
+			}
+		} else
+			qsort( users, user_count, sizeof(WhoUser *), fn );
 	}
 	default_msg_width = gdk_string_width( gkrellm_default_font(1), default_msg );
 	
@@ -277,11 +315,8 @@ static void	get_users()
 	free(str);
 }
 
-static void toggle_view() {
-	if( name_only )
-		name_only = FALSE;
-	else
-		name_only = TRUE;
+static inline void toggle_view() {
+	name_only = name_only ? FALSE : TRUE;
 }
 
 /* gtk callbacks */
@@ -357,8 +392,8 @@ void decal_text_insert( GkrellmDecal *d, gchar *s, GkrellmTextstyle *ts,
 
 #else
 /* gkrellm 2.1.11 will have these two decal text functions built in. */
-#define gkrellm_decal_text_clear	decal_text_clear
-#define gkrellm_decal_text_insert	decal_text_insert
+#define decal_text_clear	gkrellm_decal_text_clear
+#define decal_text_insert	gkrellm_decal_text_insert
 
 #endif
 
@@ -477,6 +512,8 @@ static void save_plugin_config(FILE *f)
 	fprintf(f, "%s frequency %d\n", MONITOR_CONFIG_KEYWORD, frequency);
 	fprintf(f, "%s away_time %d\n", MONITOR_CONFIG_KEYWORD, (int)away_time);
 	fprintf(f, "%s gone_time %d\n", MONITOR_CONFIG_KEYWORD, (int)gone_time);
+	fprintf(f, "%s notify_once %d\n", MONITOR_CONFIG_KEYWORD, notify_once);
+	fprintf(f, "%s notify_command %s\n", MONITOR_CONFIG_KEYWORD, notify_command);
 	for( i = 0; i < 3; i++ ) {
 		get_color_name(status_colors[i], str);
 		fprintf(f, "%s %s_color %s\n", MONITOR_CONFIG_KEYWORD, status[i], str );
@@ -511,6 +548,10 @@ static void load_plugin_config(gchar *config_line)
 		sscanf(config_data, "%d", (int *)&away_time);
 	if ( !strcmp(config_keyword, "gone_time") )
 		sscanf(config_data, "%d", (int *)&gone_time);
+	if ( !strcmp(config_keyword, "notify_once") )
+		sscanf(config_data, "%1d", &notify_once);
+	if ( !strcmp(config_keyword, "notify_command") )
+		sscanf(config_data, "%255[^\n]", notify_command);
 	for( i = 0; i < 3; i++ ) {
 		sprintf( str, "%s_color", status[i] );
 		if ( !strcmp(config_keyword, str) )
@@ -539,6 +580,8 @@ static void apply_plugin_config(void)
 	frequency =	gtk_spin_button_get_value_as_int( GTK_SPIN_BUTTON(freq_spin_button) );
 	away_time =	gtk_spin_button_get_value_as_int( GTK_SPIN_BUTTON(away_spin_button) );
 	gone_time =	gtk_spin_button_get_value_as_int( GTK_SPIN_BUTTON(gone_spin_button) );
+	notify_once = GTK_TOGGLE_BUTTON(notify_once_button)->active;
+	strncpy( notify_command, gtk_entry_get_text(GTK_ENTRY(notify_command_str)), MAX_STRING_LEN );
 	for( i = 0; i < 3; i++ ) {
 		str = strdup( gtk_entry_get_text(GTK_ENTRY(status_color_str[i])) );
 		change_color( str, &status_colors[i] );
@@ -615,6 +658,7 @@ static void create_plugin_tab(GtkWidget *tab_vbox)
 	GtkWidget	*sort_vbox;
 	GtkWidget	*idle_vbox;
 	GtkWidget	*color_hbox;
+	GtkWidget	*command_hbox;
 	GtkWidget	*label;
 	GtkWidget	*choose_button;
 	GtkWidget	*info_tab;
@@ -669,8 +713,6 @@ static void create_plugin_tab(GtkWidget *tab_vbox)
 	
 	gtk_box_pack_end( GTK_BOX(sort_vbox), sort_name_button, FALSE, FALSE, 0 );
 	gtk_box_pack_end( GTK_BOX(sort_vbox), sort_idle_button, FALSE, FALSE, 0 );
-/*	gkrellm_gtk_check_button(sort_vbox, &sort_name_button, sort_by_name,
-			FALSE, 0, "Sort users by name"); */
 
 	/* -- Idle status options tab */
 	idle_tab = gkrellm_gtk_framed_notebook_page(notebook, "Idle status");
@@ -710,6 +752,17 @@ static void create_plugin_tab(GtkWidget *tab_vbox)
 		gtk_container_add(GTK_CONTAINER(idle_vbox), color_hbox);
 	}
 	
+	command_hbox = gtk_hbox_new( FALSE, 0 );
+	gkrellm_gtk_check_button(idle_vbox, &notify_once_button, notify_once,
+			FALSE, 0, "Execute notification command once");
+	label = gtk_label_new( "Notification command:" );
+	notify_command_str = gtk_entry_new_with_max_length( 200 );
+	gtk_entry_set_text( GTK_ENTRY(notify_command_str), notify_command );
+	gtk_entry_set_editable( GTK_ENTRY(notify_command_str), TRUE);
+	gtk_box_pack_end( GTK_BOX(command_hbox), notify_command_str, FALSE, FALSE, 0);
+	gtk_box_pack_end( GTK_BOX(command_hbox), label, FALSE, FALSE, 0);
+	gtk_container_add(GTK_CONTAINER(idle_vbox), command_hbox);
+	
 	/* -- Info tab */
 	info_tab = gkrellm_gtk_framed_notebook_page(notebook, "Info");
 	info_text = gkrellm_gtk_scrolled_text_view(info_tab, NULL,
@@ -750,6 +803,8 @@ static GkrellmMonitor	plugin_mon	=
 
 GkrellmMonitor *gkrellm_init_plugin(void)
 {
+	memset( users, 0, sizeof(users) );
+	memset( cache, 0, sizeof(users) );
 	style_id = gkrellm_add_meter_style(&plugin_mon, STYLE_NAME);
 	monitor = &plugin_mon;
 	return &plugin_mon;
